from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Type

from marshy.factory.impl_marshaller_factory import get_impls

from servey_stub.errors import StubError
from servey_stub.type_definition.type_definition import TypeDefinition
from servey_stub.type_definition.type_definition_factory_abc import (
    TypeDefinitionFactoryABC,
)


def get_type_definition_factories():
    result = [i() for i in get_impls(TypeDefinitionFactoryABC)]
    result.sort(key=lambda t: t.priority)
    return result


@dataclass
class TypeDefinitionContext:
    model_dir: Path
    model_package_name: str
    type_definition_factories: List[TypeDefinitionFactoryABC] = field(
        default_factory=get_type_definition_factories
    )
    type_definitions: Dict[Type, TypeDefinition] = field(default_factory=dict)

    def get_type_definition(self, type_: Type) -> TypeDefinition:
        type_definition = self.type_definitions.get(type_)
        if type_definition:
            return type_definition
        for factory in self.type_definition_factories:
            type_definition = factory.create_type_definition(type_, self)
            if type_definition:
                self.type_definitions[type_] = type_definition
                return type_definition
        raise StubError(f"no_definition_for_type:{type_}")



import json
from dataclasses import is_dataclass, fields, MISSING
from io import StringIO
from pathlib import Path
from typing import Type, Optional

from marshy import ExternalType
from servey.util import to_snake_case

from servey_stub.type_definition.dataclass_definition import DataclassDefinition
from servey_stub.type_definition.field_definition import FieldDefinition
from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition import TypeDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.type_definition.type_definition_factory_abc import (
    TypeDefinitionFactoryABC,
)
from servey_stub.utils import update_file


class DataclassTypeDefinitionFactory(TypeDefinitionFactoryABC):
    def create_type_definition(
        self, type_: Type, context: TypeDefinitionContext
    ) -> Optional[TypeDefinition]:
        if not is_dataclass(type_):
            return
        name = type_.__name__
        snake_name = to_snake_case(name)
        model_package_name = f"{context.model_package_name}.{snake_name}.{name}"
        type_definition = TypeDefinition(
            name, ImportsDefinition([model_package_name.split(".")])
        )
        context.type_definitions[type_] = type_definition  # Prevent recursion
        model_file = Path(context.model_dir, (snake_name + ".py"))
        dataclass_definition = self.create_dataclass_definition(type_, context)
        writer = StringIO()
        dataclass_definition.write(writer)
        update_file(model_file, writer.getvalue())
        return type_definition

    @staticmethod
    def create_dataclass_definition(
        type_: Type, context: TypeDefinitionContext
    ) -> DataclassDefinition:
        imports = ImportsDefinition([["dataclasses", "dataclass"]])
        field_definitions = []
        # noinspection PyDataclass
        for field in fields(type_):
            field_type_definition = context.get_type_definition(field.type)
            if field_type_definition.imports:
                imports.add_all(field_type_definition.imports)
            schema = field.metadata.get("schemey")
            if schema and not has_ref(schema.schema):
                schema = schema.schema
                imports.add("schemey.schema_from_json")
            else:
                schema = None
            field_definition = FieldDefinition(
                name=field.name,
                type=field_type_definition.type_name,
                default_value=default_value_to_str(field),
                schema=schema,
            )
            field_definitions.append(field_definition)
        imports = imports.optimize()
        dataclass_definition = DataclassDefinition(
            name=type_.__name__,
            imports=imports,
            fields=field_definitions,
        )
        return dataclass_definition


def default_value_to_str(field):
    if field.default is MISSING:
        return None
    if field.default in (True, False, None):
        return str(field.default)
    return json.dumps(field.default)


def has_ref(schema: ExternalType) -> bool:
    if isinstance(schema, dict):
        if schema.get("$ref"):
            return True
        for s in schema.values():
            if has_ref(s):
                return True
    elif isinstance(schema, list):
        for s in schema:
            if has_ref(s):
                return True
    return False



import json
from dataclasses import dataclass
from io import IOBase
from typing import Optional

from marshy.types import ExternalItemType


@dataclass
class FieldDefinition:
    name: str
    type: str
    default_value: Optional[str] = None
    schema: Optional[ExternalItemType] = None

    def write(self, writer: IOBase):
        writer.write("    ")
        writer.write(self.name)
        writer.write(": ")
        writer.write(self.type)
        if self.schema:
            writer.write(" = field(")
            if self.default_value is not None:
                writer.write("default=")
                writer.write(self.default_value)
                writer.write(", ")
            writer.write('metadata={"schemey": schema_from_json(')
            json.dump(self.schema, writer)
            writer.write(")})")
        elif self.default_value is not None:
            writer.write(" = ")
            writer.write(self.default_value)
        writer.write("\n")



from marshy.factory.impl_marshaller_factory import register_impl
from marshy.marshaller_context import MarshallerContext

from servey_stub.type_definition.dataclass_type_definition_factory import (
    DataclassTypeDefinitionFactory,
)
from servey_stub.type_definition.enum_type_definition_factory import (
    EnumTypeDefinitionFactory,
)
from servey_stub.type_definition.generic_type_definition_factory import (
    GenericTypeDefinitionFactory,
)
from servey_stub.type_definition.primitive_type_definition_factory import (
    PrimitiveTypeDefinitionFactory,
)
from servey_stub.type_definition.type_definition_factory_abc import (
    TypeDefinitionFactoryABC,
)

priority = 100


def configure(context: MarshallerContext):
    register_impl(TypeDefinitionFactoryABC, DataclassTypeDefinitionFactory, context)
    register_impl(TypeDefinitionFactoryABC, PrimitiveTypeDefinitionFactory, context)
    register_impl(TypeDefinitionFactoryABC, EnumTypeDefinitionFactory, context)
    register_impl(TypeDefinitionFactoryABC, GenericTypeDefinitionFactory, context)



import argparse
from pathlib import Path

from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.trigger.web_trigger import WebTrigger


def main():
    parser = argparse.ArgumentParser(description="Servey")
    parser.add_argument("--mode", required=True)
    parser.add_argument("--dir", default="servey_stub")
    parser.add_argument("--only", default="")
    parser.add_argument("--exclude", default="")
    args, _ = parser.parse_known_args()
    mode = args.mode
    if mode == "lambdas":
        from servey_stub.lambda_service_definition import (
            LambdaServiceDefinition as Service,
        )
    elif mode == "http":
        from servey_stub.http_service_definition import HttpServiceDefinition as Service
    elif mode == "mock":
        from servey_stub.mock_service_definition import MockServiceDefinition as Service
    else:
        raise ValueError(f"invalid_mode:{args.mode}")
    actions = [action for action, trigger in find_actions_with_trigger_type(WebTrigger)]
    if args.only:
        only = [e.strip() for e in args.only.split(",")]
        actions = [action for action in actions if action.name in only]
    elif args.exclude:
        exclude = [e.strip() for e in args.exclude.split(",")]
        actions = [action for action in actions if action.name not in exclude]
    service = Service(actions=actions, output_dir=Path(args.dir))
    service.generate_service_code()


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from datetime import datetime
from typing import Type, Dict, Optional
from uuid import UUID

from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition import TypeDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.type_definition.type_definition_factory_abc import (
    TypeDefinitionFactoryABC,
)


def get_type_definitions() -> Dict[Type, TypeDefinition]:
    return {
        bool: TypeDefinition(bool.__name__),
        float: TypeDefinition(float.__name__),
        int: TypeDefinition(int.__name__),
        str: TypeDefinition(str.__name__),
        UUID: TypeDefinition(UUID.__name__, ImportsDefinition(["uuid.UUID"])),
        datetime: TypeDefinition(
            datetime.__name__, ImportsDefinition(["datetime.datetime"])
        ),
    }


@dataclass
class PrimitiveTypeDefinitionFactory(TypeDefinitionFactoryABC):
    priority: int = 110
    type_definitions: Dict[Type, TypeDefinition] = field(
        default_factory=get_type_definitions
    )

    def create_type_definition(
        self, type_: Type, context: TypeDefinitionContext
    ) -> Optional[TypeDefinition]:
        type_definition = self.type_definitions.get(type_)
        if type_definition:
            return type_definition



import inspect
import json
from dataclasses import dataclass
from io import IOBase
from pathlib import Path
from typing import Tuple

from marshy.types import ExternalItemType
from servey.action.action import Action

import servey_stub
from servey_stub.service_definition_abc import ServiceDefinitionABC
from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.utils import get_existing_content, update_file


@dataclass
class MockServiceDefinition(ServiceDefinitionABC):
    def generate_scaffold_pys(self, package_dir: Path, models_dir: Path):
        super().generate_scaffold_pys(package_dir, models_dir)
        update_file(
            Path(package_dir, "mock_utils.py"),
            get_existing_content(
                Path(Path(servey_stub.__file__).parent, "templates", "mock_utils.py")
            ),
        )

    def create_models(
        self, model_dir: Path, model_package_name: str
    ) -> Tuple[ImportsDefinition, TypeDefinitionContext]:
        imports, type_definition_context = super().create_models(
            model_dir, model_package_name
        )
        imports.add(["marshy"])
        imports.add([self.service_name, "mock_utils"])
        imports = imports.optimize()
        return imports, type_definition_context

    def gather_examples(self) -> ExternalItemType:
        examples = {}
        for action in self.actions:
            examples[action.name] = [
                [example.params, example.result] for example in (action.examples or [])
            ]
        return examples

    def write_class_header(self, writer: IOBase):
        writer.write("\n\nMOCKS = ")
        json.dump(self.gather_examples(), writer)
        writer.write("\n\n\n")
        writer.write("class ")
        writer.write("".join(s.title() for s in self.service_name.split("_")))
        writer.write(":\n\n")

    def write_function_body(
        self,
        action: Action,
        sig: inspect.Signature,
        context: TypeDefinitionContext,
        writer: IOBase,
    ):
        writer.write("        event_ = {\n")
        for param in sig.parameters.values():
            writer.write('            "')
            writer.write(param.name)
            writer.write('": marshy.dump(')
            writer.write(param.name)
            writer.write(", ")
            type_name = getattr(param.annotation, "__name__", None) or str(
                param.annotation
            )
            writer.write(type_name)
            writer.write("),\n")
        writer.write("        }\n")
        writer.write('        result_ = mock_utils.get_mock_result("')
        writer.write(action.name)
        writer.write('", event_, MOCKS)\n')
        writer.write(
            "        loaded_result_ = marshy.load(result_)\n        return loaded_result_\n\n"
        )



import os
from datetime import datetime
from pathlib import Path


def update_file(path: Path, content: str):
    existing_content = get_existing_content(path)
    if content != existing_content:
        with open(path, "w") as writer:
            writer.write("# GENERATED BY SERVEY-STUB\n")
            writer.write("# Manual updates to this file may be lost. \n")
            writer.write("# Updated At: ")
            writer.write(datetime.now().isoformat())
            writer.write("\n")
            writer.write(content)


def read_file(path: Path):
    with open(path, "r") as reader:
        return "".join(reader.readlines())


def get_existing_content(path: Path):
    if not os.path.exists(path):
        return
    with open(path, "r") as reader:
        lines = reader.read().split("\n")
        lines = [line for line in lines if not line.startswith("# ")]
    result = "\n".join(lines)
    return result



import inspect
import os
from dataclasses import dataclass, field
from io import IOBase
from pathlib import Path
from typing import List, Tuple

from servey.action.action import Action
from servey.trigger.web_trigger import WebTrigger

import servey_stub
from servey_stub.service_definition_abc import ServiceDefinitionABC
from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.utils import update_file, get_existing_content


@dataclass
class HttpServiceDefinition(ServiceDefinitionABC):
    install_requires: List[str] = field(default_factory=lambda: ["marshy", "requests"])
    service_root_url: str = field(default_factory=lambda: os.environ.get("SERVER_HOST"))

    def generate_scaffold_pys(self, package_dir: Path, models_dir: Path):
        super().generate_scaffold_pys(package_dir, models_dir)
        update_file(
            Path(package_dir, "http_utils.py"),
            get_existing_content(
                Path(Path(servey_stub.__file__).parent, "templates", "http_utils.py")
            ),
        )

    def create_models(
        self, model_dir: Path, model_package_name: str
    ) -> Tuple[ImportsDefinition, TypeDefinitionContext]:
        imports, type_definition_context = super().create_models(
            model_dir, model_package_name
        )
        imports.add(["marshy"])
        imports.add([self.service_name, "http_utils"])
        imports = imports.optimize()
        return imports, type_definition_context

    def write_function_body(
        self,
        action: Action,
        sig: inspect.Signature,
        context: TypeDefinitionContext,
        writer: IOBase,
    ):
        trigger = next(t for t in action.triggers if isinstance(t, WebTrigger))
        writer.write("        event_ = {\n")
        for param in sig.parameters.values():
            writer.write('            "')
            writer.write(param.name)
            writer.write('": marshy.dump(')
            writer.write(param.name)
            writer.write(", ")
            type_name = getattr(param.annotation, "__name__", None) or str(
                param.annotation
            )
            writer.write(type_name)
            writer.write("),\n")
        writer.write("        }\n")
        writer.write('        result_ = http_utils.invoke("')
        url = self.service_root_url + (
            trigger.path or ("/" + action.name.replace("_", "-"))
        )
        writer.write(url)
        writer.write('", "')
        writer.write(trigger.method.value)
        writer.write('", event_, ')
        writer.write(str(action.timeout))
        writer.write(")\n")
        writer.write(
            "        loaded_result_ = marshy.load(result_)\n        return loaded_result_\n\n"
        )



from __future__ import annotations
from dataclasses import dataclass, field
from io import IOBase
from typing import List, Union


@dataclass
class ImportsDefinition:
    imports: List[List[str]] = field(default_factory=list)

    def add(self, path: Union[str, List[str]]):
        if isinstance(path, str):
            path = path.split(".")
        self.imports.append(path)

    def add_all(self, imports: ImportsDefinition):
        self.imports.extend(imports.imports)

    def optimize(self) -> ImportsDefinition:
        root_imports = set()
        optimized_imports = {}
        for type_import in self.imports:
            key = tuple(type_import[:-1])
            values = set(s.strip() for s in type_import[-1].split(","))
            if key:
                existing = optimized_imports.get(key)
                if existing:
                    existing.update(values)
                else:
                    optimized_imports[key] = values
            else:
                root_imports.add(tuple(values))
        imports = list(list(i) for i in root_imports)
        for key, value in optimized_imports.items():
            type_import = list(key)
            value = list(value)
            value.sort()
            type_import.append(", ".join(value))
            imports.append(type_import)
        imports.sort()
        return ImportsDefinition(imports)

    def write(self, writer: IOBase):
        for type_import in self.imports:
            if len(type_import) > 1:
                writer.write("from ")
                writer.write(".".join(type_import[:-1]))
                writer.write(" ")
            writer.write("import ")
            writer.write(type_import[-1])
            writer.write("\n")



from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import servey_stub
from servey_stub.http_service_definition import HttpServiceDefinition
from tests.fixtures import ACTIONS, create_mock_file_system


class TestHttpServiceDefinition(TestCase):
    def test_generate_service_code(self):
        service = HttpServiceDefinition(
            actions=ACTIONS,
            service_name="greeter",
            service_root_url="https://greeter.com",
        )
        mock_file_system = create_mock_file_system(
            Path(Path(servey_stub.__file__).parent, "templates", "setup.txt"),
            Path(Path(servey_stub.__file__).parent, "templates", "http_utils.py"),
        )
        with (
            patch("os.path.exists", mock_file_system.exists),
            patch("builtins.open", mock_file_system.open),
            patch("pathlib.PosixPath.mkdir", mock_file_system.mkdir),
        ):
            service.generate_service_code()
        contents = {
            str(k): [s for s in v.getvalue().split("\n") if not s.startswith("#")]
            for k, v in mock_file_system.contents.items()
            if "/templates/" not in str(k)
        }
        expected_contents = {
            "servey_stub/greeter_http/greeter/__init__.py": [""],
            "servey_stub/greeter_http/greeter/models/__init__.py": [""],
            "servey_stub/greeter_http/greeter/http_utils.py": [
                "import requests",
                "from marshy.types import ExternalItemType, ExternalType",
                "",
                "",
                "def invoke(",
                "    url: str, method: str, event: ExternalItemType, timeout: int = 15",
                ") -> ExternalType:",
                '    if method.upper() in ("POST", "PUT", "PATCH"):',
                "        response = requests.request(method, url, json=event, timeout=timeout)",
                "    else:",
                "        response = requests.request(method, url, params=event, timeout=timeout)",
                "    return response.json()",
                "",
            ],
            "servey_stub/greeter_http/greeter/models/number_stats.py": [
                "from dataclasses import dataclass",
                "",
                "",
                "@dataclass",
                "class NumberStats:",
                "    name: str",
                "    value: int",
                "",
            ],
            "servey_stub/greeter_http/greeter/greeter.py": [
                "from greeter import http_utils",
                "from greeter.models.number_stats import NumberStats",
                "import marshy",
                "",
                "",
                "class Greeter:",
                "",
                '    def say_hello(self, name: str = "User"):',
                "        event_ = {",
                '            "name": marshy.dump(name, str),',
                "        }",
                '        result_ = http_utils.invoke("https://greeter.com/say-hello", "get", event_, 15)',
                "        loaded_result_ = marshy.load(result_)",
                "        return loaded_result_",
                "",
                "",
                "    def save_number_stats(self, number_stats: NumberStats, insert: bool = False):",
                "        event_ = {",
                '            "number_stats": marshy.dump(number_stats, NumberStats),',
                '            "insert": marshy.dump(insert, bool),',
                "        }",
                '        result_ = http_utils.invoke("https://greeter.com/save-number-stats", "post", event_, 15)',
                "        loaded_result_ = marshy.load(result_)",
                "        return loaded_result_",
                "",
                "",
                "",
            ],
            "servey_stub/greeter_http/setup.py": [
                "import setuptools",
                "",
                "setuptools.setup(",
                '    name="greeter",',
                '    version="0.1.0",',
                '    packages=setuptools.find_packages(exclude=("tests*")),',
                "    install_requires=['marshy', 'requests'],",
                ")",
                "",
            ],
        }
        self.assertEqual(expected_contents, contents)



import inspect
import json
from dataclasses import dataclass
from enum import Enum
from io import StringIO, IOBase
from typing import Type, Optional

from servey.util import to_snake_case

from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition import TypeDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.type_definition.type_definition_factory_abc import (
    TypeDefinitionFactoryABC,
)
from servey_stub.utils import update_file


@dataclass
class EnumTypeDefinitionFactory(TypeDefinitionFactoryABC):
    priority: int = 110

    def create_type_definition(
        self, type_: Type, context: TypeDefinitionContext
    ) -> Optional[TypeDefinition]:
        if not inspect.isclass(type_) or not issubclass(type_, Enum):
            return
        name = type_.__name__
        type_definition = TypeDefinition(
            name,
            ImportsDefinition(
                [context.model_package_name + "." + to_snake_case(name) + "." + name]
            ),
        )
        model_file = context.model_dir / (to_snake_case(name) + ".py")
        writer = StringIO()
        self.write_enum_definition(type_, writer)
        update_file(model_file, writer.getvalue())
        return type_definition

    @staticmethod
    def write_enum_definition(type_: Type, writer: IOBase):
        writer.write("from enum import Enum\n\nclass ")
        writer.write(type_.__name__)
        writer.write("(Enum):\n")
        for e in type_:
            writer.write("    ")
            writer.write(e.name)
            writer.write(" = ")
            writer.write(json.dumps(e.value))
            writer.write("\n")



from typing import List, Tuple, Dict, Optional, Callable

from marshy.types import ExternalItemType, ExternalType

from servey_stub.errors import StubError


def get_mock_result(
    action_name: str,
    event: ExternalItemType,
    mocks: Dict[str, List[Tuple[ExternalItemType, ExternalType]]],
):
    custom_handler = get_custom_handler(action_name)
    if custom_handler:
        result = custom_handler(event)
        return result
    action_mocks = mocks.get(action_name) or []
    for params, result in action_mocks:
        if event == params:
            return result
    raise StubError("no_mock_for_event")


def get_custom_handler(
    action_name: str,
) -> Optional[Callable[[ExternalItemType], ExternalType]]:
    try:
        # noinspection PyUnresolvedReferences
        from . import custom_mocks

        return getattr(custom_mocks, action_name)
    except ModuleNotFoundError:
        return None



from abc import ABC, abstractmethod
from typing import Type, Optional

from servey_stub.type_definition.type_definition import TypeDefinition

_TypeDefinitionContext = (
    "servey_stub.type_definition.type_definition_context.TypeDefinitionContext"
)


class TypeDefinitionFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create_type_definition(
        self, type_: Type, context: _TypeDefinitionContext
    ) -> Optional[TypeDefinition]:
        """Generate a type name and return it, (and generate any additional required model files)"""



from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import servey_stub
from servey_stub.lambda_service_definition import LambdaServiceDefinition
from tests.fixtures import ACTIONS, create_mock_file_system


class TestLambdaServiceDefinition(TestCase):
    def test_generate_service_code(self):
        service = LambdaServiceDefinition(actions=ACTIONS, service_name="greeter")
        mock_file_system = create_mock_file_system(
            Path(Path(servey_stub.__file__).parent, "templates", "setup.txt"),
            Path(Path(servey_stub.__file__).parent, "templates", "lambda_utils.py"),
        )
        with (
            patch("os.path.exists", mock_file_system.exists),
            patch("builtins.open", mock_file_system.open),
            patch("pathlib.PosixPath.mkdir", mock_file_system.mkdir),
        ):
            service.generate_service_code()
        contents = {
            str(k): [s for s in v.getvalue().split("\n") if not s.startswith("#")]
            for k, v in mock_file_system.contents.items()
            if "/templates/" not in str(k)
        }
        expected_contents = {
            "servey_stub/greeter_lambda/greeter/__init__.py": [""],
            "servey_stub/greeter_lambda/greeter/models/__init__.py": [""],
            "servey_stub/greeter_lambda/greeter/lambda_utils.py": [
                "import json",
                "import typing",
                "import boto3",
                "from marshy.types import ExternalType",
                "",
                "_lambda_client = None",
                "",
                "",
                "def get_lambda_client():",
                "    # pylint: disable=W0603",
                "    global _lambda_client",
                "    if not _lambda_client:",
                '        _lambda_client = boto3.client("lambda")',
                "    return _lambda_client",
                "",
                "",
                "def invoke(",
                '    function_name: str, event: typing.Dict, invocation_type: str = "RequestResponse"',
                ") -> ExternalType:",
                "    request_payload = json.dumps(event)",
                "    response = get_lambda_client().invoke(",
                "        FunctionName=function_name,",
                "        InvocationType=invocation_type,",
                "        Payload=request_payload,",
                "    )",
                '    response_payload = response.get("Payload")',
                "    if response_payload:",
                "        data = json.load(response_payload)",
                "        return data",
                "",
            ],
            "servey_stub/greeter_lambda/greeter/models/number_stats.py": [
                "from dataclasses import dataclass",
                "",
                "",
                "@dataclass",
                "class NumberStats:",
                "    name: str",
                "    value: int",
                "",
            ],
            "servey_stub/greeter_lambda/greeter/greeter.py": [
                "from greeter import lambda_utils",
                "from greeter.models.number_stats import NumberStats",
                "import marshy",
                "",
                "",
                "class Greeter:",
                "",
                '    def say_hello(self, name: str = "User"):',
                "        event_ = {",
                '            "action_name": "say_hello",',
                '            "params": {',
                '                "name": marshy.dump(name, str),',
                "            }",
                "        }",
                '        result_ = lambda_utils.invoke_lambda("ServeyMain-dev-say_hello", event_)',
                "        loaded_result_ = marshy.load(str, result_)",
                "        return loaded_result_",
                "",
                "",
                "    def save_number_stats(self, number_stats: NumberStats, insert: bool = False):",
                "        event_ = {",
                '            "action_name": "save_number_stats",',
                '            "params": {',
                '                "number_stats": marshy.dump(number_stats, NumberStats),',
                '                "insert": marshy.dump(insert, bool),',
                "            }",
                "        }",
                '        result_ = lambda_utils.invoke_lambda("ServeyMain-dev-save_number_stats", event_)',
                "        loaded_result_ = marshy.load(int, result_)",
                "        return loaded_result_",
                "",
                "",
                "",
            ],
            "servey_stub/greeter_lambda/setup.py": [
                "import setuptools",
                "",
                "setuptools.setup(",
                '    name="greeter",',
                '    version="0.1.0",',
                '    packages=setuptools.find_packages(exclude=("tests*")),',
                "    install_requires=['boto3', 'marshy'],",
                ")",
                "",
            ],
        }
        self.assertEqual(expected_contents, contents)



import requests
from marshy.types import ExternalItemType, ExternalType


def invoke(
    url: str, method: str, event: ExternalItemType, timeout: int = 15
) -> ExternalType:
    if method.upper() in ("POST", "PUT", "PATCH"):
        response = requests.request(method, url, json=event, timeout=timeout)
    else:
        response = requests.request(method, url, params=event, timeout=timeout)
    return response.json()



import inspect
import json
import os
from abc import ABC, abstractmethod
from dataclasses import field, dataclass
from io import StringIO, IOBase
from pathlib import Path
from typing import List, Optional, Tuple

from servey.action.action import Action
from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.trigger.web_trigger import WebTrigger
from servey.util import to_snake_case, get_servey_main

import servey_stub
from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.utils import update_file, read_file


@dataclass
class ServiceDefinitionABC(ABC):
    actions: List[Action] = field(
        default_factory=lambda: [
            action for action, trigger in find_actions_with_trigger_type(WebTrigger)
        ]
    )
    service_name: str = field(default_factory=get_servey_main)
    output_dir: Path = Path("servey_stub")
    description: Optional[str] = field(
        default_factory=lambda: os.environ.get("SERVEY_API_DESCRIPTION")
    )
    version: str = field(
        default_factory=lambda: os.environ.get("SERVEY_API_VERSION") or "0.1.0"
    )
    install_requires: List[str] = field(default_factory=lambda: ["marshy"])

    @property
    def service_type_name(self):
        return "".join(
            to_snake_case(self.__class__.__name__).split("_service_definition")
        )

    def create_models(
        self, model_dir: Path, model_package_name: str
    ) -> Tuple[ImportsDefinition, TypeDefinitionContext]:
        imports = ImportsDefinition()
        context = TypeDefinitionContext(
            model_dir=model_dir, model_package_name=model_package_name
        )
        for type_ in _get_types_for_actions(self.actions):
            type_definition = context.get_type_definition(type_)
            if type_definition.imports:
                imports.add_all(type_definition.imports)
        imports = imports.optimize()
        return imports, context

    def generate_service_code(self):
        service_name = to_snake_case(self.service_name)
        project_dir, package_dir, models_dir = self.generate_directories(service_name)
        self.generate_scaffold_pys(package_dir, models_dir)
        imports, context = self.create_models(models_dir, service_name + ".models")
        self.generate_service_py(package_dir, service_name, imports, context)
        self.generate_setup_py(project_dir, service_name)

    def generate_directories(self, service_name: str) -> Tuple[Path, Path, Path]:
        project_dir = Path(self.output_dir, service_name + "_" + self.service_type_name)
        package_dir = Path(project_dir, service_name)
        models_dir = Path(package_dir, "models")
        models_dir.mkdir(parents=True, exist_ok=True)
        return project_dir, package_dir, models_dir

    def generate_scaffold_pys(self, package_dir: Path, models_dir: Path):
        update_file(Path(package_dir, "__init__.py"), "")
        update_file(Path(models_dir, "__init__.py"), "")

    def generate_setup_py(self, project_dir: Path, service_name: str):
        setup_content = read_file(
            Path(Path(servey_stub.__file__).parent, "templates", "setup.txt")
        )
        setup_content = setup_content.format(
            service_name=service_name,
            version=self.version,
            install_requires=self.install_requires,
        )
        update_file(Path(project_dir, "setup.py"), setup_content)

    def generate_service_py(
        self,
        package_dir: str,
        service_name: str,
        imports: ImportsDefinition,
        context: TypeDefinitionContext,
    ):
        writer = StringIO()
        self.write(imports, context, writer)
        service_content = writer.getvalue()
        service_file = Path(package_dir, service_name + ".py")
        update_file(service_file, service_content)

    def write(
        self, imports: ImportsDefinition, context: TypeDefinitionContext, writer: IOBase
    ):
        imports.write(writer)
        self.write_class_header(writer)
        for action in self.actions:
            self.write_action(action, context, writer)

    def write_class_header(self, writer: IOBase):
        writer.write("\n\n")
        writer.write("class ")
        writer.write("".join(s.title() for s in self.service_name.split("_")))
        writer.write(":\n\n")

    def write_action(
        self, action: Action, context: TypeDefinitionContext, writer: IOBase
    ):
        sig = inspect.signature(action.fn)
        self.write_function_header(action, sig, context, writer)
        self.write_function_body(action, sig, context, writer)
        writer.write("\n")

    @staticmethod
    def write_function_header(
        action: Action,
        sig: inspect.Signature,
        context: TypeDefinitionContext,
        writer: IOBase,
    ):
        writer.write("    def ")
        writer.write(action.name)
        writer.write("(self")
        for param in sig.parameters.values():
            writer.write(", ")
            writer.write(param.name)
            writer.write(": ")
            writer.write(context.type_definitions[param.annotation].type_name)
            if param.default is not sig.empty:
                writer.write(" = ")
                if isinstance(param.default, str):
                    writer.write(json.dumps(param.default))
                else:
                    writer.write(str(param.default))
        writer.write("):\n")

    @abstractmethod
    def write_function_body(
        self,
        action: Action,
        sig: inspect.Signature,
        context: TypeDefinitionContext,
        writer: IOBase,
    ):
        """Write the body of a function"""


def _get_types_for_actions(actions: List[Action]):
    for action in actions:
        sig = inspect.signature(action.fn)
        for param in sig.parameters.values():
            yield param.annotation
        yield sig.return_annotation



from dataclasses import dataclass
from typing import Optional

from servey_stub.type_definition.imports_definition import ImportsDefinition


@dataclass
class TypeDefinition:
    type_name: str
    imports: Optional[ImportsDefinition] = None



from dataclasses import dataclass, field
from io import IOBase
from typing import Optional, List

from servey_stub.type_definition.field_definition import FieldDefinition
from servey_stub.type_definition.imports_definition import ImportsDefinition


@dataclass
class DataclassDefinition:
    """
    Definition for code generation for a dataclass to be used as a data transfer object
    """

    name: str
    imports: ImportsDefinition
    fields: List[FieldDefinition] = field(default_factory=list)
    description: Optional[str] = None

    def write(self, writer: IOBase):
        self.imports.write(writer)
        writer.write("\n\n@dataclass\nclass ")
        writer.write(self.name)
        writer.write(":\n")
        if self.description:
            writer.write('    """')
            writer.write(self.description.replace("\n", "\n    "))
            writer.write('    """\n')
        for field_ in self.fields:
            field_.write(writer)



from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import servey_stub
from servey_stub.mock_service_definition import MockServiceDefinition
from tests.fixtures import ACTIONS, create_mock_file_system


class TestMockServiceDefinition(TestCase):
    def test_generate_service_code(self):
        service = MockServiceDefinition(actions=ACTIONS, service_name="greeter")
        mock_file_system = create_mock_file_system(
            Path(Path(servey_stub.__file__).parent, "templates", "setup.txt"),
            Path(Path(servey_stub.__file__).parent, "templates", "mock_utils.py"),
        )
        with (
            patch("os.path.exists", mock_file_system.exists),
            patch("builtins.open", mock_file_system.open),
            patch("pathlib.PosixPath.mkdir", mock_file_system.mkdir),
        ):
            service.generate_service_code()
        contents = {
            str(k): [s for s in v.getvalue().split("\n") if not s.startswith("#")]
            for k, v in mock_file_system.contents.items()
            if "/templates/" not in str(k)
        }
        expected_contents = {
            "servey_stub/greeter_mock/greeter/__init__.py": [""],
            "servey_stub/greeter_mock/greeter/models/__init__.py": [""],
            "servey_stub/greeter_mock/greeter/mock_utils.py": [
                "from typing import List, Tuple, Dict, Optional, Callable",
                "",
                "from marshy.types import ExternalItemType, ExternalType",
                "",
                "from servey_stub.errors import StubError",
                "",
                "",
                "def get_mock_result(",
                "    action_name: str,",
                "    event: ExternalItemType,",
                "    mocks: Dict[str, List[Tuple[ExternalItemType, ExternalType]]],",
                "):",
                "    custom_handler = get_custom_handler(action_name)",
                "    if custom_handler:",
                "        result = custom_handler(event)",
                "        return result",
                "    action_mocks = mocks.get(action_name) or []",
                "    for params, result in action_mocks:",
                "        if event == params:",
                "            return result",
                '    raise StubError("no_mock_for_event")',
                "",
                "",
                "def get_custom_handler(",
                "    action_name: str,",
                ") -> Optional[Callable[[ExternalItemType], ExternalType]]:",
                "    try:",
                "        # noinspection PyUnresolvedReferences",
                "        from . import custom_mocks",
                "",
                "        return getattr(custom_mocks, action_name)",
                "    except ModuleNotFoundError:",
                "        return None",
                "",
            ],
            "servey_stub/greeter_mock/greeter/models/number_stats.py": [
                "from dataclasses import dataclass",
                "",
                "",
                "@dataclass",
                "class NumberStats:",
                "    name: str",
                "    value: int",
                "",
            ],
            "servey_stub/greeter_mock/greeter/greeter.py": [
                "from greeter import mock_utils",
                "from greeter.models.number_stats import NumberStats",
                "import marshy",
                "",
                "",
                'MOCKS = {"say_hello": [[{"name": "Dev"}, "Hello Dev!"]], "save_number_stats": [[{"number_stats": {"name": "one", "value": 1}}, 1]]}',
                "",
                "",
                "class Greeter:",
                "",
                '    def say_hello(self, name: str = "User"):',
                "        event_ = {",
                '            "name": marshy.dump(name, str),',
                "        }",
                '        result_ = mock_utils.get_mock_result("say_hello", event_, MOCKS)',
                "        loaded_result_ = marshy.load(result_)",
                "        return loaded_result_",
                "",
                "",
                "    def save_number_stats(self, number_stats: NumberStats, insert: bool = False):",
                "        event_ = {",
                '            "number_stats": marshy.dump(number_stats, NumberStats),',
                '            "insert": marshy.dump(insert, bool),',
                "        }",
                '        result_ = mock_utils.get_mock_result("save_number_stats", event_, MOCKS)',
                "        loaded_result_ = marshy.load(result_)",
                "        return loaded_result_",
                "",
                "",
                "",
            ],
            "servey_stub/greeter_mock/setup.py": [
                "import setuptools",
                "",
                "setuptools.setup(",
                '    name="greeter",',
                '    version="0.1.0",',
                '    packages=setuptools.find_packages(exclude=("tests*")),',
                "    install_requires=['marshy'],",
                ")",
                "",
            ],
        }
        self.assertEqual(expected_contents, contents)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

extras_require = {
    "dev": [
        "black~=23.3",
        "boto3~=1.26",
        "requests~=2.31",
        "pytest~=7.2",
        "pytest-cov~=4.0",
        "pytest-xdist~=3.2",
        "pylint~=2.17",
    ],
}
extras_require["all"] = list(
    {
        dependency
        for dependencies in extras_require.values()
        for dependency in dependencies
    }
)


setuptools.setup(
    name="servey-stub",
    author="Tim O'Farrell",
    author_email="tofarr@gmail.com",
    description="A service stub generator for servey",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tofarr/servey-stub",
    packages=setuptools.find_packages(exclude=("tests*", "examples")),
    package_data={"": ["*.txt"]},
    include_package_data=True,
    install_requires=[
        "servey~=2.8",
    ],
    extras_require=extras_require,
    setup_requires=["setuptools-git-versioning"],
    setuptools_git_versioning={"enabled": True, "dirty_template": "{tag}"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import json
import typing
import boto3
from marshy.types import ExternalType

_lambda_client = None


def get_lambda_client():
    # pylint: disable=W0603
    global _lambda_client
    if not _lambda_client:
        _lambda_client = boto3.client("lambda")
    return _lambda_client


def invoke(
    function_name: str, event: typing.Dict, invocation_type: str = "RequestResponse"
) -> ExternalType:
    request_payload = json.dumps(event)
    response = get_lambda_client().invoke(
        FunctionName=function_name,
        InvocationType=invocation_type,
        Payload=request_payload,
    )
    response_payload = response.get("Payload")
    if response_payload:
        data = json.load(response_payload)
        return data



import inspect
import os
from dataclasses import dataclass, field
from io import IOBase
from pathlib import Path
from typing import List, Tuple

from servey.action.action import Action
from servey.util import get_servey_main

import servey_stub
from servey_stub.service_definition_abc import ServiceDefinitionABC
from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.utils import get_existing_content, update_file


def _default_lambda_name_pattern() -> str:
    servey_main = get_servey_main().title().replace("_", "")
    return servey_main + "-dev-{fn_name}"


@dataclass
class LambdaServiceDefinition(ServiceDefinitionABC):
    lambda_name_pattern: str = field(default_factory=_default_lambda_name_pattern)
    use_router_for_all: bool = field(
        default_factory=lambda: int(os.environ.get("SERVEY_AWS_ROUTER_FOR_ALL", "0"))
        == 1
    )
    router_name: str = "servey_router"
    install_requires: List[str] = field(default_factory=lambda: ["boto3", "marshy"])

    def generate_scaffold_pys(self, package_dir: Path, models_dir: Path):
        super().generate_scaffold_pys(package_dir, models_dir)
        update_file(
            Path(package_dir, "lambda_utils.py"),
            get_existing_content(
                Path(Path(servey_stub.__file__).parent, "templates", "lambda_utils.py")
            ),
        )

    def create_models(
        self, model_dir: Path, model_package_name: str
    ) -> Tuple[ImportsDefinition, TypeDefinitionContext]:
        imports, type_definition_context = super().create_models(
            model_dir, model_package_name
        )
        imports.add(["marshy"])
        imports.add([self.service_name, "lambda_utils"])
        imports = imports.optimize()
        return imports, type_definition_context

    def write_function_body(
        self,
        action: Action,
        sig: inspect.Signature,
        context: TypeDefinitionContext,
        writer: IOBase,
    ):
        writer.write('        event_ = {\n            "action_name": "')
        writer.write(action.name)
        writer.write('",\n            "params": {\n')
        for param in sig.parameters.values():
            writer.write('                "')
            writer.write(param.name)
            writer.write('": marshy.dump(')
            writer.write(param.name)
            writer.write(", ")
            type_name = getattr(param.annotation, "__name__", None) or str(
                param.annotation
            )
            writer.write(type_name)
            writer.write("),\n")
        writer.write("            }\n        }\n")
        writer.write('        result_ = lambda_utils.invoke_lambda("')
        use_router = self.use_router_for_all or "<locals>" in action.fn.__qualname__
        fn_name = self.router_name if use_router else action.name
        writer.write(self.lambda_name_pattern.format(fn_name=fn_name))
        writer.write('", event_)\n')
        result_type_name = getattr(sig.return_annotation, "__name__", None) or str(
            sig.return_annotation
        )
        writer.write("        loaded_result_ = marshy.load(")
        writer.write(result_type_name)
        writer.write(", result_)\n        return loaded_result_\n\n")



from typing import Type, Optional

import typing_inspect

from servey_stub.type_definition.imports_definition import ImportsDefinition
from servey_stub.type_definition.type_definition import TypeDefinition
from servey_stub.type_definition.type_definition_context import TypeDefinitionContext
from servey_stub.type_definition.type_definition_factory_abc import (
    TypeDefinitionFactoryABC,
)


class GenericTypeDefinitionFactory(TypeDefinitionFactoryABC):
    def create_type_definition(
        self, type_: Type, context: TypeDefinitionContext
    ) -> Optional[TypeDefinition]:
        origin = typing_inspect.get_origin(type_)
        if not origin:
            return
        name = str(type_)
        imports = ImportsDefinition()
        if name.startswith("typing."):
            prefix = name.split("[", maxsplit=1)[0]
            imports.add(prefix)
            arg_names = []
            for arg in typing_inspect.get_args(type_):
                if arg is not type(None):
                    arg_type_definition = context.get_type_definition(arg)
                    arg_names.append(arg_type_definition.type_name)
                    if arg_type_definition.imports:
                        imports.add_all(arg_type_definition.imports)
            type_name = f"{prefix}[{', '.join(arg_names)}]"
            return TypeDefinition(type_name, imports)
